/*
 * Created on May 11, 2005
 *
 */
package com.sunlabs.archsim.build;

import java.io.File;
import java.io.IOException;
import java.util.Set;
import java.util.Stack;

import org.xml.sax.*;
import org.xml.sax.helpers.DefaultHandler;

import com.sunlabs.archsim.base.common.Parameters;
import com.sunlabs.archsim.classRegistry.ClassRegistry;
import com.sunlabs.archsim.engine.Director;

import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;

/**
 *
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 * 
 * @author ib27688
 *
 * Load the XML file with the description of the network that is 
 * to be modeled, and assemble the model.
 */
class ModelLoader extends DefaultHandler {

	public static final String MODEL= "model";
	public static final String COMPONENT= "component";
	public static final String CONNECTION= "connection";
	public static final String FROM= "from";
	public static final String TO= "to";
	public static final String NAME= "name";	
	public static final String VALUE= "value";
	public static final String ACTIVE= "active";
	public static final String TYPE= "type";
	public static final String PARAMETER= "parameter";
	public static final String TERMINAL= "terminal";
	public static final String COMMENT= "comment";
	
	/** registru where to register component classes */
	private ClassRegistry registry;
	
	/** director */
	private Director director= Director.getInstance();
	
	/** the context of the current element */
	private Stack<String> context;
	
	/** the current component record */
	private ComponentRecord cmpRecord;
	/** the current connection record */
	private ConnectionRecord connRecord;
	/** the current connection end */
	private ConnectionPoint connPoint;
	/** the current set of attributes */
	private Parameters attributes;
	
	/** the collection of components and connections */
	private ModelDescription model;
	
	/** Locator provided by the parser */
	private Locator locator;
	
	public ModelLoader(ModelDescription m) {
		model= m;
		registry= ClassRegistry.instance;
		context= new Stack<String>();
		// prepare empty records
		cmpRecord= new ComponentRecord();
		connRecord= new ConnectionRecord();
		connPoint= new ConnectionPoint();
		attributes= new Parameters();
	} // ModelLoader constructor
	
	
	
	/** load components form a file */
	public void loadFile(String fileName) {
	       	SAXParserFactory factory = SAXParserFactory.newInstance();
	        factory.setValidating(true);
	        //factory.setNamespaceAware(true);
	        try {
	            // Parse the input
	            SAXParser saxParser = factory.newSAXParser();
	            saxParser.parse( new File(fileName), this);

	        } catch (SAXParseException spe) {
	           // Error generated by the parser
	           director.error("\n** Parsing error"
	              + ", line " + spe.getLineNumber()
	              + ", uri " + spe.getSystemId()
	              + "  " + spe.getMessage() );

	           // Use the contained exception, if any
	           Exception  x = spe;
	           if (spe.getException() != null)
	               x = spe.getException();
	           x.printStackTrace();
	           System.exit(1);

	        } catch (SAXException sxe) {
	            // Error generated by this application
	            // (or a parser-initialization error)
	      
	        		// first take a look at a message associated with
	        		// 
	        		String message= sxe.getMessage();
	        		if (!message.equals("")) {
	        			director.error(message);
	        		}
	        		
	        		Exception  x = sxe;
	        		if (sxe.getException() != null)
	            		x = sxe.getException();
	        		x.printStackTrace();
	        		System.exit(1);

	        } catch (ParserConfigurationException pce) {
	            	// Parser with specified options can't be built
	            	pce.printStackTrace();

	        } catch (IOException ioe) {
	        		// I/O error
	        		ioe.printStackTrace();
	        }		
	} // loadFile
		
	/** 
	 * The callback that provides a Locator - used to get line numbers
	 * for error messages, for example.
	 */
    public void setDocumentLocator(Locator l) {
    		locator = l;
    }

	/**
	 *  prep work to be done at the beginning of the document 
	 */
	public void startDocument() throws SAXException {
		// What do we do?
		// we have not entered the "root" element - list
		context.clear();
		
	} // startDocument
	
	/** 
	 * Finishing work to be done at the end of the XML file
	 */
	public void endDocument() throws SAXException {
		// What do we do?
		// nothing at the moment
	} // endDocument

	public void startElement (
			String namespaceURI,
			String lName, // local name
			String qName, // qualified name
			Attributes attrs) 
	throws SAXException {
		String eName = lName; // element name
	     if ("".equals(eName)) eName = qName; // namespaceAware = false
	     // establish the context
	     // note: the parser checks that the document is well-formed
	     // and valid
	     context.push(new String(eName));
	     
	     /** record source code lines and model name */
	     if (eName.equals(MODEL)) {
	     	// get the model name
	     	// the parser will ensure the attibute is there
	     	// pass through attributes to guard agains a future change
	     	if (attrs != null) {
	     		for (int i= 0; i< attrs.getLength(); i++) {
	     			String attrName= attrs.getLocalName(i);
	     			if ("".equals(attrName)) attrName = attrs.getQName(i);
	     			if (attrName.equals(NAME))
	     				model.name= new String(attrs.getValue(i).trim());
	     		} // for
	     	} // if
	     }
	     else if (eName.equals(COMPONENT)) {
	     	cmpRecord.line= locator.getLineNumber();
	     	// get the component name and type
	     	// the parser will ensure the attibute is there
	     	// pass through attributes to guard against a future change
	     	if (attrs != null) {
	     		for (int i= 0; i< attrs.getLength(); i++) {
	     			String attrName= attrs.getLocalName(i);
	     			if ("".equals(attrName)) attrName = attrs.getQName(i);
	     			if (attrName.equals(NAME)) {
	     				cmpRecord.name= new String(attrs.getValue(i).trim());
	     			}
	     			else if (attrName.equals(TYPE)) {
	     				// does this class exist in the registry?
	     				String className= attrs.getValue(i);
	     				Class c= registry.getComponentType(className);
	     				if (c != null) {
	     					// class found, record it
	     					cmpRecord.type= c;
	     				}
	     				else {
	     					// class not found, error
	     					throw new SAXException("Error in line " 
	     							+ locator.getLineNumber()
	     							+ ": Class tag \"" + className 
	     							+ "\" not recognized");
	     				}     				
	     			}
	     			else if (attrName.equals(PARAMETER)) {
	     				// record the parameter in the current attibutes
	     				attributes.add(PARAMETER, new String(attrs.getValue(i).trim()));	     				
	     				// relic
	     				// cmpRecord.parameter= new String(attrs.getValue(i).trim());
	     			}
	     		} // for
	     	} // if	     	
	     }
	     else if (eName.equals(CONNECTION)) {
	     	// record the source line
	     	connRecord.line= locator.getLineNumber();
	     	// get the connection name
	     	// the parser will ensure the attibute is there
	     	// pass through attributes to guard against a future change
	     	if (attrs != null) {
	     		for (int i= 0; i< attrs.getLength(); i++) {
	     			String attrName= attrs.getLocalName(i);
	     			if ("".equals(attrName)) attrName = attrs.getQName(i);
	     			if (attrName.equals(NAME)) {
	     				connRecord.name= new String(attrs.getValue(i).trim());
	     			}
	     			else if (attrName.equals(ACTIVE)) {
	     				connRecord.dataActive= true;
	     			}
	     			else if (attrName.equals(VALUE)) {
	     				connRecord.dataActive= true;
	     				connRecord.dataValue= new String(attrs.getValue(i).trim());
	     			}
	     		} // for
	     	} // if	     	
	     }
	     else if (eName.equals(FROM) || eName.equals(TO)) {
	     	// record the source line
	     	connPoint.line= locator.getLineNumber();
	     	// get the connection data - component and terminal
	     	// the parser will ensure the attibute is there
	     	// pass through attributes to guard against a future change
	     	if (attrs != null) {
	     		for (int i= 0; i< attrs.getLength(); i++) {
	     			String attrName= attrs.getLocalName(i);
	     			if ("".equals(attrName)) attrName = attrs.getQName(i);
	     			if (attrName.equals(COMPONENT)) {
	     				// record the component name
	     				connPoint.componentName= new String(attrs.getValue(i).trim());
	     			} 
	     			else if (attrName.equals(TERMINAL)) {
	     				// record the terminal name
	     				connPoint.terminalName= new String(attrs.getValue(i).trim());	     				
	     			}
	     		} // for
	     		
				if (!model.componentNameMap.containsKey(connPoint.componentName)) {
					throw new SAXException("Error in line "
							+ locator.getLineNumber()
						+ ": component "
						+ connPoint.componentName
						+ " was not defined, therefore"
						+ " the connection has an unkown end.");
				}
	     		    		
	    			// add a conection point to the conection record
	    			Set<ConnectionPoint> endPoints= eName.equals(FROM)? 
	    					connRecord.sources : connRecord.destinations;
	    			boolean check= endPoints.add(connPoint);
	    			if (!check) {
	    				// connection specified twice    				
	    				throw new SAXException("Error in line " 
	    						+ connPoint.line
	    						+ ": The end point of the connection"
							+ " was specified twice.");
	    			}
	    			// reset the connection point
	    			// prepare for a subsequent one
	    			connPoint= new ConnectionPoint();
	     	} // if	attrs != null     	
	     }
	     else if (eName.equals(PARAMETER)) {
	    	 // store the line number for error messages
	    	 int line= locator.getLineNumber();
	    	 // prepare for 
	    	 String name= null;
	    	 String value= null;
	    	 if (attrs != null) {
	     		for (int i= 0; i< attrs.getLength(); i++) {
	     			String attrName= attrs.getLocalName(i);
	     			if ("".equals(attrName)) attrName = attrs.getQName(i);
	     			if (attrName.equals(NAME)) {
	     				// record the name of the attribute
	     				name= new String(attrs.getValue(i).trim());
	     			} 
	     			else if (attrName.equals(VALUE)) {
	     				// record the name value of the attribute
	     				value= new String(attrs.getValue(i).trim());	     				
	     			}
	     		} // for
	    	 } // if attrs != null
	    	 
	     	if (name != null && value != null) {
	    		 attributes.add(name, value);
	    	}
	    	else {
				throw new SAXException("Error in line " 
						+ line
						+ ": Either the name or the value of a parameter not provided. Name: "
						+ name 
						+ " Value: "
						+ value);
	    	}
	     } // if eName==PARAMETER
	} // startElement
	
    public void endElement(String namespaceURI,
            String lName, // simple name
            String qName  // qualified name
           )
    throws SAXException {
    		// leave the current context
    		String ctx= (String) context.pop();
    		if (ctx.equals(COMPONENT)) {
    			// record current attributes
    			cmpRecord.attributes= attributes;
    			// prepare for the next set of attributes 
    			attributes= new Parameters();
       			// record the component in the model
    			model.componentNameMap.put(cmpRecord.name, cmpRecord);
    			// reset the component record
    			// get ready for a subsequent component
    			cmpRecord= new ComponentRecord();
    		}
    		else if (ctx.equals(CONNECTION)) {
    			// record existing parameters
    			connRecord.attributes= attributes;
    			// prepare for next set of parameters
    			attributes= new Parameters();
    			// record the connection
    			model.connectionRecordList.add(connRecord);
    			// reset the connection record
    			// get ready for a subsequent connection
    			connRecord= new ConnectionRecord();
    		}
    		else if (ctx.equals(FROM)) {
    		}
    		else if (ctx.equals(TO)) {	
    		}
    		else if (ctx.equals(MODEL)) {
    		}
    } // endElement
    
	public void characters(char buf[], int offset, int len)
	throws SAXException {
		// ignore all character data
	} // characters
	
    // treat validation errors as fatal
    public void error(SAXParseException e)
    throws SAXParseException
    {
        throw e;
    } // error

    // dump warnings too
    public void warning(SAXParseException err)
    throws SAXParseException
    {
        director.error("** Warning"
        		+ ", line " + err.getLineNumber()
            	+ ", uri " + err.getSystemId()
        		+ "   " + err.getMessage());
    } // warning

    
    public static void main(String[] args) {
		ClassRegistry reg=  ClassRegistry.instance;
		reg.load("testRegLoader.xml");
		
		
		ModelDescription model= new ModelDescription();
		ModelLoader modelLoader= new ModelLoader(model);
		modelLoader.loadFile("Example.xml");
		
		model.print();
    	
    } // main
	
	
} // LoadModel