/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

/*
* Created on Mar 17, 2005
*
*/
package com.sunlabs.archsim.gasp;

import java.util.Iterator;
import java.util.StringTokenizer;
// import java.util.Vector;

import com.sunlabs.archsim.base.hierarchy.CompositeEntity;
import com.sunlabs.archsim.base.simulation.Command;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;

/**
* @author ib27688
*
* An action that sits at the heart of the branch group.
*/
abstract public class BranchAction extends Action {

//default terminal names
public static final String IN= "In";
public static final String ACKIN= "ackIn";
public static final String OUT= "Out";
public static final String ACKOUT= "ackOut";
public static final String DELIMITER= "_";

/** the number of branches */
protected int numBranches;

/** the input for the data */
protected InputTerminal in;
/* acknowldegmnet that the input has been received */
protected OutputTerminal ackIn;
        
/** outputs for the data */
OutputTerminal[] out;
/** acknowldegment that the output has been received */
InputTerminal[] ackOut;

protected boolean inPending;
protected boolean[] outPending;
protected int branchSelected;

/** data received through in terminal */
protected Object dataReceived;

/** Delay to copy the data.
        */
protected Delay copyDelay;

/** 
 * The number of branches defaults to two.
 * When the number of branches is unknown at the time of construction,
 * the number of branches can be later set by invoking setNumBranches.
 * Not that the setup must be done before any terminals are attached to the action,
 * in particular out and ackOut terminals.
 *  
 * @param name component name
 */
/** default is two branches */
protected BranchAction(String name) {
        this(name, 2);
}

/** provide the name and the number of branches */
protected BranchAction(String name, int branches) {
        super(name);
        
        // set data structures for the terminals
        numBranches= branches;
        in= null;
        ackIn= null;
        out= new OutputTerminal[numBranches];
        ackOut= new InputTerminal[numBranches];
        for (int i= 0; i< numBranches; i++) {
                out[i]= null;
                ackOut[i]= null;
        }
        
        // initialize the flags
        inPending= false;
        outPending= new boolean[numBranches];
        for (int i= 0; i< numBranches; i++) outPending[i]= false;
        
        
        dataReceived= null;
        copyDelay= new Delay(0); // no extra delay to copy
        
        notifyStateIsActiveCommand= new StateActiveCommand();
        notifyStateIsInactiveCommand= new StateInactiveCommand();
} // Branch

/** allow to set the group right away */
protected BranchAction(String name, int branches, CompositeEntity g) {
        this(name, branches);
        g.addMember(this);
}


public void setNumBranches(int n) {
    // set data structures for the terminals
    numBranches= n;

    // prepare the array of output terminals
    out= new OutputTerminal[numBranches];
    ackOut= new InputTerminal[numBranches];
    for (int i= 0; i< numBranches; i++) {
            out[i]= null;
            ackOut[i]= null;
    }
    
    // initialize the flags
    inPending= false;
    outPending= new boolean[numBranches];
    for (int i= 0; i< numBranches; i++) outPending[i]= false;
}

public int getNumBranches() {
	return numBranches;
}

/* Attach an individual terminal. Name restrictions:
        * BranchAction.IN, BranchAction.ACKIN,
        * BranckAction.OUT + BranchAction.DELIMITER+number
        * BranckAction.ACKOUT + BranchAction.DELIMITER+number
        * where the number is between 0 and numBranches.
        * 
        * Use attachIN and attachOut to avoid naming restrictions.
        */
public void attach(Terminal t) {
        StringTokenizer st= new StringTokenizer(t.getBaseName(), DELIMITER);
        String prefix;
        String number;
        boolean twice= false;
        boolean illegalName= false;
        try {
                // get the first part of the terminal name
                prefix= st.nextToken();
                // check whether the name and direction are OK
                if (t.getBaseName() == IN && t instanceof InputTerminal) {
                        // we have an input
                        if (terminals.add(t)) {
                                t.attachAction(this);
                                this.in= (InputTerminal)t;
                        } else 
                                twice= true;                 
                } else if (t.getBaseName() == ACKIN && t instanceof OutputTerminal) {
                        if (terminals.add(t)) {
                                t.attachAction(this);
                                this.ackIn= (OutputTerminal)t;
                        } else 
                                twice= true;                                 
                } else if (st.countTokens()==1) {
                        number= st.nextToken();
                        // convert string to an integer, exception if bad format
                        int branch= Integer.decode(number).intValue();
                        if (0 > branch || branch >= numBranches) {
                                illegalName= true;
                        } else if (terminals.contains(t)) {
                                twice= true;
                        }
                        else if  (prefix == OUT && t instanceof OutputTerminal) {
                                out[branch]= (OutputTerminal)t;
                        }
                        else if (prefix == ACKOUT && t instanceof InputTerminal) {
                                ackOut[branch]= (InputTerminal)t;
                        }
                        else {
                                // the terminal name was no good after all
                                illegalName= true;
                        }
                }
                else {
                        // prefix was no good, or wrong name structure
                        illegalName= true;
                }
        } // try
        catch (Exception e) {
                illegalName= true;
        }

        if (illegalName) {
                director.error("BranchAction " + getAlias() 
                		+ " AKA " + getName()
                		+ ", id= "
					+ getID() + ": attempt to attach a terminal with"
                     	+ " illegal name, " 
					+ t.getName()
                     	+ ", or illegal direction/role " + t.getRole());                 
        }
        else if (twice) {
                director.error("BranchAction " + getAlias() 
                		+ " AKA " + getName()
                		+ ", id= "
					+ getID() + ": attempt to attach the same terminal"
                     + " twice: " + t.getName());                         
        }
        
} // Attach
        

/** attach the terminals on the input side */
public void attachIn(InputTerminal in, OutputTerminal ackIn) {
        // add terminal to the set
        if (in != null && ackIn != null &&
                        !terminals.contains(in) && !terminals.contains(ackIn)) {
                terminals.add(in);
                ((Terminal)in).attachAction(this);
                this.in= in;
        
                terminals.add(ackIn);
                ((Terminal)ackIn).attachAction(this);
                this.ackIn= ackIn;
        }
        else {
                if (in == null) {
                        director.error("BranchAction " + getAlias() + ", id= "
                                        + getID() + ": null in terminal attached" );                         
                }
                else if (terminals.contains(in)){
                        director.error("BranchAction " + getAlias() + ", id= "
                                        + getID() + ": terminal attached twice: "
                                        + in.getAlias() + ", id= " + in.getID());                         
                }
                if (ackIn == null) {
                        director.error("BranchAction " + getAlias() + ", id= "
                                        + getID() + ": null ackIn terminal attached" );                                                         
                }
                else if (terminals.contains(ackIn)) {
                        director.error("BranchAction " + getAlias() + ", id= "
                                + getID() + ": terminal attached twice: "
                                + ackIn.getAlias() + ", id= " + ackIn.getID());
                }
        }
} // attachIn

public void attachOut(OutputTerminal out, InputTerminal ackOut, int n ) {
        if (n >=0 && n < numBranches
                && out != null && ackOut != null
                && !terminals.contains(out) && !terminals.contains(ackOut)) {
                        this.out[n]= out;
                        this.ackOut[n]= ackOut;
                        // add the terminals to the set 
                        terminals.add(out); 
                        terminals.add(ackOut);
                        
                        // attach this action to the terminals
                        ((Terminal)this.out[n]).attachAction(this);
                        ((Terminal)this.ackOut[n]).attachAction(this);
        }
        else {
                if (n<0 || n >= numBranches) {
                        director.error("BranchAction " + getAlias() + ", id= "
                                        + getID() + ": illegal branch number used: "
                                        + n + "; should be between 0 and less than " 
                                        + numBranches );                                         
                }
                else {
                        if (out == null) {
                                director.error("BranchAction " + getAlias() + ", id= "
                                                + getID() + "null out terminal attached "
                                                + "at branch " + n);                         
                        }
                        else if (terminals.contains(out)){
                                director.error("BranchAction " + getAlias() + ", id= "
                                                + getID() + ": terminal attached twice: "
                                                + out.getAlias() + ", id= " + out.getID());                         
                        }
                        if (ackOut == null) {
                                director.error("BranchAction " + getAlias() + ", id= "
                                                + getID() + "null ackOut terminal attached "
                                                + "at branch " + n);                                                         
                        }
                        else if (terminals.contains(ackOut)) {
                                director.error("BranchAction " + getAlias() + ", id= "
                                        + getID() + ": terminal attached twice: "
                                        + ackOut.getAlias() + ", id= " + ackOut.getID());
                        }
                }
        }
        
} // attachOut
        
/** 
        * This method is used to select the branch where the
        * data is passed.
        * 
        * @param data data received on the input
        * @return the index of the branch selected
        */
public abstract int selectBranch(Object data);

/**
        * Select the data that needs to be copied. By default this is 
        * the straight copy.
        * 
        * @param data input data
        * @return output data
        */
public Object copyData(Object data) {
        return data;
} // copyData

protected class StateActiveCommand extends Command {
        public void execute(Object param) {
                Terminal src= (Terminal)param;
                if (journalAction) {
                        State state= src.getState();
                        director.journal("Action " + getAlias()
                                        + ", id= " + getID() + " notifed that state "
                                        + state.getAlias() + ", id= " + state.getID() 
                                        + " is active");                                         
                } 
                // has a new input arrived?
                if (src instanceof InputTerminal) {
                        if (src == in) {
                                inPending= true;
                                dataReceived= in.getData();
                                branchSelected= selectBranch(dataReceived);
                                if (outPending[branchSelected]) {
//                                        director.schedule(fireCommand, 
//                                                        copyData(dataReceived), copyDelay );
                                        computeAndFire(branchSelected);
                                }
                        }
                        else { // src == ackOut[i]
                                int branch=-1;
                                // find the branch
                                for (int i= 0; i< numBranches; i++) {
                                        if (src == ackOut[i]) {
                                                branch= i;
                                                break;
                                        }
                                } // for
                                if (branch >=0 && branch < numBranches) {
	                                // recod that this branch can fire
	                                outPending[branch] = true;
	                                if (inPending && branchSelected== branch) {
	                                        computeAndFire(branchSelected);
	                                }
                                }
                                else {
                                	director.fatalError("BranchAction"
                                			+ BranchAction.this.getAlias() 
										+ " , id= " 
										+ BranchAction.this.getID()
										+ ", received input on an unrecognized port "
										+ src.getAlias() + ", id= " + src.getID());
                                }
                        } // else
                } // if
        } // execute
} // StateActiveCommand

/*
        * Compute the result and maka appropriae inputs inactive and
        * appropriate outputs active;
        * 
        * @param i index of the selected branch
        */
protected void computeAndFire(int i) {
        // build the data structure required by fireCommand
        TerminalValuePair pair= new TerminalValuePair();
        Terminal t= out[i];
        Object value= copyData(dataReceived);
        pair.terminal= t;
        pair.value= value;
        // make "outputs" active
        
        director.schedule(fireCommand, pair, copyDelay );
        ackIn.setStateActive(null);
        // make relevant "inputs" inactive
        in.setStateInactive();
        ackOut[i].setStateInactive();

        // reset the flags
        inPending= false;
        outPending[branchSelected]= false;
} // computeAndFire


protected class StateInactiveCommand extends Command {
        public void execute(Object param) {
                Terminal src= (Terminal)param;
                if (journalAction) {
                        State state= src.getState();
                        director.journal("Action " + getAlias()
                                        + ", id= " + getID() + " notifed that state "
                                        + state.getAlias() + ", id= " + state.getID() 
                                        + " is INactive");                                         
                } // if (journalAction)

                // record that an input is inactive
                // that is, reset the corresponding "pending" flag
                // QUESTION: what is the consequence for a firing that might be in progress?
                // That is, what is the consequence for firings that have been scheduled?
                // Consider rewriting fireCommand so that that command resets the
                // "pending" flags
                if (src instanceof InputTerminal) {
                    if (src == in) {
                            inPending= false;
                    }
                    else { // src == ackOut[i]
                            int branch=-1;
                            // find the branch
                            for (int i= 0; i< numBranches; i++) {
                                    if (src == ackOut[i]) {
                                            branch= i;
                                            break;
                                    }
                            } // for
                            if (branch >=0 && branch < numBranches) {
                                // recod that this branch cannot fire
                                outPending[branch] = false;
                            }
                            else {
                            	director.fatalError("BranchAction"
                            			+ BranchAction.this.getAlias() 
									+ " , id= " 
									+ BranchAction.this.getID()
									+ ", received input on an unrecognized port "
									+ src.getAlias() + ", id= " + src.getID());
                            }
                    } // else
            } // if                
                
        }         
} // StateInactiveCommand

public boolean selfCheck() {
        // first do the default self check
        boolean pass= true;

        // check connections
        Iterator ti= terminals.iterator();
        while (pass && ti.hasNext()) {
                Terminal t= (Terminal)(ti.next());
                pass= pass && t.isAttached(this);
                if (!pass) {
                        director.error("Action " + getAlias() + ", id= " + getID() 
                                        + ", thinks it is attached to terminal " + t.getAlias()
                                        + ", id= " + t.getID() + ", but the terminal is"
                                        + " not attached to the action");
                }
        }
        
        pass= pass && (in != null);
        if (!pass) {
                director.error("BranchAction " + getAlias() + ", id= "
                                + getID() + ", does not have the in terminal set.");
        }
        
        pass= pass && (ackIn != null);
        if (!pass) {
                director.error("BranchAction " + getAlias() + ", id= "
                                + getID() + ", does not have the ack-in terminal set.");
        }                 
                
        // make sure all outputs are set
        for (int i= 0; i< out.length; i++) {
                pass= pass && (out[i]!= null);
                if (!pass) {
                        director.error("BranchAction " + getAlias() + ", id= "
                                        + getID() + ", does not have output terminal " + 
                                        i + " set.");
                }
        }
        
        // make sure all ackOuts are set
        for (int i= 0; i< out.length; i++) {
                pass= pass && (ackOut[i]!= null);
                if (!pass) {
                        director.error("BranchAction " + getAlias() + ", id= "
                                        + getID() + ", does not have output terminal " + 
                                        i + " set.");
                }
        }

        return pass;         
} // selfCheck
} // class BranchAction

