package edu.berkeley.sbp;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.Sequence.Position;
import edu.berkeley.sbp.util.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;

/** an efficient representation of a collection of trees (Tomita's shared packed parse forest) */
public abstract class Forest<T> /*extends PrintableTree<Forest.MyBody<T>>*/
    implements Visitable<Forest.Body<T>>,
               IntegerMappable,
               GraphViz.ToGraphViz {

    private static int master_idx = 0;
    private final int idx = master_idx++;
    public int toInt() { return idx; }

    public abstract void expand(TaskList tl, HashSet<Tree<T>> ht);
    public abstract void gather(TaskList tl, HashSet<Tree<T>>[] ht, HashSet<Tree<T>> target);

    public static class TaskList extends ArrayList<TaskList.Task> {
        public interface Task {
            public void perform();
        }
        public class ExpandTask implements Task {
            private Forest f;
            private HashSet hs;
            public ExpandTask(Forest f, HashSet hs) { this.f = f; this.hs = hs; }
            public void perform() { f.expand(TaskList.this, hs); }
        }
        public class GatherTask implements Task {
            private Forest f;
            private HashSet[] ht;
            private HashSet hs;
            public GatherTask(Forest f, HashSet<Tree<?>>[] ht, HashSet<Tree<?>> hs) { this.f = f; this.hs = hs; this.ht = ht;}
            public void perform() { f.gather(TaskList.this, ht, hs); }
        }
        public void expand(Forest f, HashSet hs) {
            add(new ExpandTask(f, hs));
        }
        public void gather(Forest f, HashSet[] ht, HashSet hs) {
            add(new GatherTask(f, ht, hs));
        }
        public void run() {
            while(true) {
                if (isEmpty()) return;
                Task task = get(size()-1);
                remove(size()-1);
                task.perform();
            }
        }
    }

    /** assume that this forest contains exactly one tree and return it; otherwise throw an exception */
    public final Tree<T> expand1() throws Ambiguous, ParseFailed {
        try {
            Iterator<Tree<T>> it = expand(true).iterator();
            if (!it.hasNext()) throw new ParseFailed();
            return it.next();
        } catch (InnerAmbiguous ia) { throw new Ambiguous(ia.f); }
    }

    /** expand this forest into a set of trees */
    public HashSet<Tree<T>> expand(boolean toss) {
        /*
        final HashSetTreeConsumer<T> ret = new HashSetTreeConsumer<T>();
        visit(new TreeMaker2<T>(toss, ret), null, null);
        */
        TaskList tl = new TaskList();
        HashSet<Tree<T>> ret = new HashSet<Tree<T>>();
        tl.expand(this, ret);
        tl.run();

        if (toss && ret.size() > 1) throw new InnerAmbiguous(this);
        return ret;
    }

    private static class InnerAmbiguous extends RuntimeException {
        public final Forest<?> f;
        public InnerAmbiguous(Forest<?> f) { this.f = f; }
    }

    static interface TreeConsumer<T> {
        public void addTree(Tree<T> t);
    }
    static class HashSetTreeConsumer<T> extends HashSet<Tree<T>> implements TreeConsumer<T> {
        public void addTree(Tree<T> t) {
            super.add(t);
        }
    }

    static        <T> Forest<T> singleton(Input.Location loc, Position p) {
        return create(loc, null, new Forest[] { }, new Object[0], false, true, p); }
    static        <T> Forest<T> singleton(Input.Location loc, Forest<T> body, Position p) {
        //return create(loc, null, new Forest[] { body },  false, true, p);
        return body;
    }
    static        <T> Forest<T> leaf(Input.Location loc, T tag, Position p) { return create(loc, tag, null, null, false, false, p); }
    public static <T> Forest<T> create(Input.Location loc, T tag, Forest<T>[] tokens, Object[] labels, boolean unwrap, boolean singleton, Position p) {
        return new MyBody<T>(loc, tag, tokens, labels, unwrap, singleton, p);
    }
    // Body //////////////////////////////////////////////////////////////////////////////

    protected static interface Body<T> extends GraphViz.ToGraphViz {
        void expand(int i, TreeMaker<T> h);
    }
    public abstract void edges(GraphViz.Node n);
    public boolean ambiguous() { return false; }
    protected static class MyBody<T> extends Forest<T> implements Body<T> /* extends PrintableTree<Forest<T>> implements */ {

        public boolean isTransparent() { return false; }
        public boolean isHidden() { return false; }
        public GraphViz.Node toGraphViz(GraphViz gv) {
            if (gv.hasNode(this)) return gv.createNode(this);
            GraphViz.Node n = gv.createNode(this);
            n.label = headToString()==null?"":headToString();
            n.directed = true;
            n.comment = reduction==null?null:reduction+"";
            edges(n);
            return n;
        }
        boolean edges = false;
        public void edges(GraphViz.Node n) {
            if (edges) return;
            edges = true;
            for(int i=0; i<tokens.length; i++) {
                if (i==tokens.length-1 && unwrap && !tokens[i].ambiguous()) {
                    tokens[i].edges(n);
                } else {
                    n.edge(tokens[i], labels==null?null:labels[i]);
                }
            }
        }

        public <B,C> void visit(Invokable<Forest.Body<T>,B,C> ivbc, B b, C c) {
            ivbc.invoke(this, b, c);
        }

        private final Input.Location    location;
        private final T                 tag;
        private final Forest<T>[]       tokens;
        private final Object[]          labels;
        private final boolean           unwrap;
        private final boolean           singleton;
        private final Sequence.Position reduction;

        private MyBody(Input.Location loc, T tag, Forest<T>[] tokens, Object[] labels, boolean unwrap, boolean singleton, Position reduction) {
            this.location = loc;
            this.tag = tag;
            this.tokens = tokens==null ? emptyForestArray : new Forest[tokens.length];
            if (tokens != null) System.arraycopy(tokens, 0, this.tokens, 0, tokens.length);
            if (tokens != null) for(int i=0; i<tokens.length; i++) if (tokens[i]==null) throw new Error(i+"");
            this.unwrap = unwrap;
            this.singleton = singleton;
            this.reduction = reduction;
            this.labels = labels;
        }
        public void gather(TaskList tl, HashSet<Tree<T>>[] ht, HashSet<Tree<T>> target) {
            gather(tl, ht, target, new Tree[ht.length], 0);
        }
        private void gather(TaskList tl, HashSet<Tree<T>>[] ht, HashSet<Tree<T>> target, Tree[] trees, int i) {
            if (i==ht.length) {
                target.add(new Tree<T>(null, tag, trees));
                return;
            }
            for(Tree<T> tree : ht[i]) {
                if (unwrap && i==trees.length-1) {
                    // I think this is wrong
                    Tree[] trees2 = new Tree[trees.length - 1 + tree.numChildren()];
                    System.arraycopy(trees, 0, trees2, 0, trees.length-1);
                    for(int j=0; j<tree.numChildren(); j++)
                        trees2[trees.length-1+j] = tree.child(j);
                    target.add(new Tree<T>(null, tag, trees2));
                } else {
                    trees[i] = tree;
                    gather(tl, ht, target, trees, i+1);
                    trees[i] = null;
                }
            }
        }
        public void expand(TaskList tl, HashSet<Tree<T>> ht) {
            if (singleton) {
                tokens[0].expand(tl, ht);
                return;
            }
            HashSet<Tree<T>>[] children = new HashSet[tokens.length];
            tl.gather(this, children, ht);
            for(int i=0; i<children.length; i++) {
                children[i] = new HashSet<Tree<T>>();
                tl.expand(tokens[i], children[i]);
            }
        }

        public void expand(final int i, final TreeMaker<T> h) {
            if (singleton) {
                tokens[0].visit(h, null, i);
                return;
            }
            if (i==0) h.start(tag, location);

            if (i==tokens.length) {
                h.finish(tag, location);

            } else if (unwrap && i==tokens.length-1) {
                if (tokens[i] != null)
                    tokens[i].visit(h, null, 0);

            } else {
                tokens[i].visit(new TreeMaker<T>(h.toss) {
                    public void start(T head, Input.Location loc) { }
                    public void addTree(Tree<T> t, Object label) { toks.add(t); labs.add(label); }
                    public void finish(T head, Input.Location loc) {
                        int old = h.toks.size();
                        h.addTree(new Tree<T>(loc, head, toks.toArray(tree_hint), labs.toArray(string_hint)), labels==null?null:labels[i]);
                        expand(i+1, h);
                        while(h.toks.size() > old) h.toks.remove(h.toks.size()-1);
                        while(h.labs.size() > old) h.labs.remove(h.labs.size()-1);
                    }
                }, null, null);
            }
        }

        protected String  headToString()         { return tag==null?null:tag.toString(); }
        protected String  headToJava()           { return "null"; }
        protected String  left()                 { return "{"; }
        protected String  right()                { return "}"; }
        protected boolean ignoreSingleton()      { return false; }
    }


    // Ref //////////////////////////////////////////////////////////////////////////////

    /**
     *  This class represents a partially complete collection of
     *  forests to be viewed as a forest at some later date; once
     *  viewed, it becomes immutable
     */
    static class Ref<T> extends Forest<T> {
        public void expand(TaskList tl, HashSet<Tree<T>> ht) {
            for (Forest<T> f : hp) f.expand(tl, ht);
        }
        public void gather(TaskList tl, HashSet<Tree<T>>[] ht, HashSet<Tree<T>> target) {
            throw new Error();
        }
        public HashSet<GSS.Phase.Node> parents = new HashSet<GSS.Phase.Node>();
        public boolean contains(Forest f) {
            return hp.contains(f);
        }
        public boolean ambiguous() {
            if (hp.size()==0) return false;
            if (hp.size()==1) return hp.iterator().next().ambiguous();
            return true;
        }
        private FastSet<Forest<T>> hp = new FastSet<Forest<T>>();
        public Ref() { }
        public int toInt() {
            if (hp.size()==1) return hp.iterator().next().toInt();
            return super.toInt();
        }
        public void merge(Forest p) { if (p!=this) hp.add(p, true); }

        public boolean isTransparent() { return hp.size()==1; }
        public boolean isHidden() { return hp.size()==0; }
        public void edges(GraphViz.Node n) {
            if (hp.size()==1) { hp.iterator().next().edges(n); return; }
            for(Forest f : hp) f.edges(n);
        }
        public GraphViz.Node toGraphViz(GraphViz gv) {
            //if (hp.size()==0) return null;
            if (hp.size()==1) return hp.iterator().next().toGraphViz(gv);
            if (gv.hasNode(this)) return gv.createNode(this);
            GraphViz.Node n = gv.createNode(this);
            n.label = "?";
            n.color = "red";
            for(Forest f : hp) n.edge(f, null);
            return n;
        }

        public <B,C> void visit(Invokable<Forest.Body<T>,B,C> ivbc, B b, C c) {
            if (hp==null) return;
            for(Forest<T> f : hp)
                f.visit(ivbc, b, c);
        }
        public Forest resolve() { return this; }
    }

    public abstract <B,C> void visit(Invokable<Forest.Body<T>,B,C> ivbc, B b, C c);
    private static class TreeMaker2<T> extends TreeMaker<T> {
        private TreeConsumer<T> tc;
        public TreeMaker2(boolean toss, TreeConsumer<T> tc) { super(toss); this.tc = tc; }
        public void finish(T head, Input.Location loc) { tc.addTree(new Tree<T>(loc, head, toks.toArray(tree_hint), labs.toArray(string_hint)));; }
        public void start(T head, Input.Location loc) { }
        public void addTree(Tree<T> t, Object label) { toks.add(t); labs.add(label); }
    }
    private static abstract class TreeMaker<T> implements Invokable<Forest.Body<T>,Boolean,Integer>/*, TreeConsumer<T>*/ {
        public ArrayList<Tree<T>> toks = new ArrayList<Tree<T>>();
        public ArrayList<Object>  labs = new ArrayList<Object>();
        private boolean toss;
        protected T head;
        public TreeMaker(boolean toss) { this.toss = toss; }
        public abstract void start(T head, Input.Location loc);
        public abstract void finish(T head, Input.Location loc);
        public abstract void addTree(Tree<T> t, Object label);
        public void invoke(Forest.Body<T> bod, Boolean o, Integer i) {
            if (i==null) {
                ArrayList<Tree<T>> toks = this.toks;
                this.toks = new ArrayList<Tree<T>>();
                ArrayList<Object> labs = this.labs;
                this.labs = new ArrayList<Object>();
                bod.expand(0, this);
                this.toks = toks;
                this.labs = labs;
            } else {
                bod.expand(i, this);
            }
        }
    }

    // Statics //////////////////////////////////////////////////////////////////////////////

    private static Tree[] tree_hint = new Tree[0];
    private static String[] string_hint = new String[0];
    private static final Forest[] emptyForestArray = new Forest[0];

    protected String  headToString()    { return null; }
    protected String  headToJava()      { return "null"; }
    protected String  left()            { return "<?"; }
    protected String  right()           { return "?>"; }
    protected boolean ignoreSingleton() { return true; }
}
